# pylint: disable=invalid-name
from __future__ import absolute_import

import uuid
from dataclasses import field
from typing import Optional, List, ClassVar, Type
from datetime import datetime

from marshmallow import Schema, fields
from marshmallow_dataclass import dataclass, NewType

from jwthenticator.consts import JWT_ALGORITHM, JWT_ALGORITHM_FAMILY

# Define the UUID type that uses Marshmallow's UUID + Python's UUID
UUID = NewType("UUID", uuid.UUID, field=fields.UUID)


# Data dataclasses (that match the sqlalchemy models)
@dataclass  # pylint: disable=used-before-assignment
class KeyData:
    Schema: ClassVar[Type[Schema]] = Schema
    id: int
    created: datetime
    expires_at: datetime
    key_hash: str
    key: Optional[str] = field(default=None, repr=False, metadata=dict(load_only=True))

    async def is_valid(self) -> bool:
        return self.expires_at > datetime.utcnow()


@dataclass
class RefreshTokenData:
    Schema: ClassVar[Type[Schema]] = Schema
    id: int
    created: datetime
    expires_at: datetime
    token: str
    key_id: int

    async def is_valid(self) -> bool:
        return self.expires_at > datetime.utcnow()


@dataclass
class JWTPayloadData:
    Schema: ClassVar[Type[Schema]] = Schema
    token_id: UUID   # JWT token identifier
    identifier: UUID # Machine the JWT was issued to identifier
    iat: int    # Issued at timestamp
    exp: int    # Expires at timestamp

    async def is_valid(self) -> bool:
        return self.exp > datetime.utcnow().timestamp()


# Request dataclasses
@dataclass
class KeyRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    key: str


@dataclass
class RegisterKeyRequest(KeyRequest):
    identifier: UUID # Machine the JWT was issued to identifier


@dataclass
class RefreshRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    refresh_token: str
    identifier: UUID


@dataclass
class JWTValidateRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    jwt: str


@dataclass
class AuthRequest:
    Schema: ClassVar[Type[Schema]] = Schema
    key: str = field(repr=False)
    identifier: UUID


# Response dataclasses
@dataclass
class TokenResponse:
    Schema: ClassVar[Type[Schema]] = Schema
    jwt: str
    refresh_token: Optional[str] = None


@dataclass
class BoolResponse:
    Schema: ClassVar[Type[Schema]] = Schema
    result: Optional[bool]
    message: Optional[str] = None


@dataclass
class JWKSResponse:
    """
    See https://auth0.com/docs/tokens/references/jwks-properties
    """
    Schema: ClassVar[Type[Schema]] = Schema
    x5c: List[str]
    n: bytes
    e: bytes
    alg: str = JWT_ALGORITHM
    kty: str = JWT_ALGORITHM_FAMILY
    use: str = "sig"



from __future__ import absolute_import

from typing import Tuple, Optional
from urllib.parse import urlparse

from Cryptodome.PublicKey import RSA

from jwthenticator.consts import RSA_KEY_STRENGTH, RSA_PUBLIC_KEY, RSA_PRIVATE_KEY, RSA_PUBLIC_KEY_PATH, RSA_PRIVATE_KEY_PATH


def get_rsa_key_pair() -> Tuple[str, Optional[str]]:
    """
    Get RSA key pair.
    Will try to get them by this order:
    1. RSA_PUBLIC/PRIVATE_KEY_PATH
    2. RSA_PUBLIC/PRIVATE_KEY
    3. Generate new keys.
    :return (public_key, private_key): A key pair tuple.
        Will raise exception if key paths are given and fail to read.
    """
    if RSA_PUBLIC_KEY_PATH is not None:
        # Read public key.
        with open(RSA_PUBLIC_KEY_PATH) as f_obj:
            public_key = f_obj.read()

        # Read private key if given.
        private_key = None
        if RSA_PRIVATE_KEY_PATH is not None:
            with open(RSA_PRIVATE_KEY_PATH) as f_obj:
                private_key = f_obj.read()

        return (public_key, private_key)

    if RSA_PUBLIC_KEY is not None:
        return (RSA_PUBLIC_KEY, RSA_PRIVATE_KEY)

    return create_rsa_key_pair()


def create_rsa_key_pair() -> Tuple[str, str]:
    """
    Create RSA key pair.
    Function is sync so it can be used from __init__ func.
    :return (public_key, private_key): A key pair tuple.
    """
    key = RSA.generate(RSA_KEY_STRENGTH)
    public_key = key.publickey().export_key().decode()
    private_key = key.export_key().decode()
    return public_key, private_key


def verify_url(url: str) -> bool:
    """
    Verifies a given URL is valid and will be accepted by http client.
    """
    parsed_url = urlparse(url)
    return all([parsed_url.scheme, parsed_url.netloc])


def fix_url_path(url: str) -> str:
    """
    Add "/" to end of URL, if URL has path and doesn't end with "/"
    the path will be removed by urljoin.
    """
    return url if url.endswith("/") else url + "/"



from __future__ import absolute_import

from os import environ

DAYS_TO_SECONDS = lambda x: x * 60 * 24


# Server consts
PORT = int(environ.get("PORT", 8080))
URL_PREFIX = URL_PREFIX = environ.get('URL_PREFIX', '')
DISABLE_EXTERNAL_API = bool(environ.get("DISABLE_EXTERNAL_API", ""))
DISABLE_INTERNAL_API = bool(environ.get("DISABLE_INTERNAL_API", ""))

# JWT consts
JWT_ALGORITHM = environ.get("JWT_ALGORITHM", "RS256")
JWT_ALGORITHM_FAMILY = environ.get("JWT_ALGORITHM_FAMILY", "RSA")
JWT_LEASE_TIME = int(environ.get("JWT_LEASE_TIME", 30 * 60)) # In seconds - 30 minutes
RSA_KEY_STRENGTH = int(environ.get("RSA_KEY_STRENGTH", 2048))

# Token consts
KEY_EXPIRY = int(environ.get("KEY_EXPIRY", DAYS_TO_SECONDS(120)))  # In seconds
REFRESH_TOKEN_EXPIRY = int(environ.get("REFRESH_TOKEN_EXPIRY", DAYS_TO_SECONDS(60)))   # In seconds

# Encryption keys
RSA_PUBLIC_KEY = environ.get("RSA_PUBLIC_KEY", None)
RSA_PRIVATE_KEY = environ.get("RSA_PRIVATE_KEY", None)
# Keys from file
RSA_PUBLIC_KEY_PATH = environ.get("RSA_PUBLIC_KEY_PATH", None)
RSA_PRIVATE_KEY_PATH = environ.get("RSA_PRIVATE_KEY_PATH", None)

# DB consts
DB_CONNECTOR = environ.get("DB_CONNECTOR", "postgresql+pg8000")
DB_USER = environ.get("DB_USER", "postgres")
DB_PASS = environ.get("DB_PASS", "")
DB_HOST = environ.get("DB_HOST", "localhost")
DB_NAME = environ.get("DB_NAME", "jwthenticator")

DB_URI = environ.get("DB_URI", f"{DB_CONNECTOR}://{DB_USER}:{DB_PASS}@{DB_HOST}/{DB_NAME}")



from __future__ import absolute_import

from typing import Optional, Tuple
from hashlib import sha512

from Cryptodome.PublicKey import RSA
from jwt.utils import force_unicode, to_base64url_uint

from jwthenticator import schemas
from jwthenticator.tokens import TokenManager
from jwthenticator.keys import KeyManager
from jwthenticator.consts import JWT_ALGORITHM, JWT_ALGORITHM_FAMILY, JWT_LEASE_TIME
from jwthenticator.utils import get_rsa_key_pair
from jwthenticator.exceptions import ExpiredError

INT_TO_BIG_ENDIAN = lambda x: force_unicode(to_base64url_uint(x))


class JWThenticatorAPI:
    """
    High level authentication API functions.
    Class uses TokenManager + KeyManager and provides API logic using them.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, rsa_key_pair: Tuple[str, Optional[str]] = get_rsa_key_pair(), jwt_lease_time: int = JWT_LEASE_TIME,
                 jwt_algorithm: str = JWT_ALGORITHM, jwt_algorithm_family: str = JWT_ALGORITHM_FAMILY):
        """
        Class can be initiated without giving any parameter, will generate RSA key pair by itself.
        :param rsa_key_pair: (public_key, private_key) RSA key pair. Will generate keys if not given
            keys or paths through consts. In canse private_key is None will only support token validation.
        """
        self.public_key, self._private_key = rsa_key_pair
        self.jwt_algorithm = jwt_algorithm
        self.jwt_algorithm_family = jwt_algorithm_family
        self.jwt_lease_time = jwt_lease_time

        self.token_manager = TokenManager(self.public_key, self._private_key, self.jwt_algorithm, self.jwt_lease_time)
        self.key_manager = KeyManager()


    async def authenticate(self, request: schemas.AuthRequest) -> schemas.TokenResponse:
        """
        Authenticate using a key.
        :return: access + refresh tokens if successfull
        """
        key_hash = sha512(request.key.encode()).hexdigest()
        key_obj = await self.key_manager.get_key(key_hash)
        if not await key_obj.is_valid():
            raise ExpiredError("Key is expired.")

        jwt_token = await self.token_manager.create_access_token(request.identifier)
        refresh_token = await self.token_manager.create_refresh_token(key_obj.id)

        return schemas.TokenResponse(
            jwt=jwt_token,
            refresh_token=refresh_token
        )


    async def refresh(self, request: schemas.RefreshRequest) -> schemas.TokenResponse:
        """
        Use a refresh token to create a new access token.
        """
        refresh_token_obj = await self.token_manager.load_refresh_token(request.refresh_token)
        if not await refresh_token_obj.is_valid():
            raise ExpiredError("Token is expired.")

        jwt_token = await self.token_manager.create_access_token(request.identifier)
        return schemas.TokenResponse(jwt=jwt_token)


    async def validate(self, request: schemas.JWTValidateRequest) -> schemas.BoolResponse:
        """
        Validate a JWT token.
        Raises a %s exception if token is invalid.
        """
        token = await self.token_manager.load_access_token(request.jwt)
        if not await token.is_valid():
            raise ExpiredError("Token is expired.")
        return schemas.BoolResponse(result=True)


    async def register_key(self, request: schemas.RegisterKeyRequest) -> schemas.BoolResponse:
        """
        Register a new key so it can be used for authentication.
        """
        await self.key_manager.create_key(request.key, request.identifier)
        return schemas.BoolResponse(result=True)


    async def is_key_registerd(self, request: schemas.KeyRequest) -> schemas.BoolResponse:
        """
        Check if a given key is already registered.
        """
        key_hash = sha512(request.key.encode()).hexdigest()
        result = await self.key_manager.check_key_exists(key_hash)
        return schemas.BoolResponse(result=result)


    async def get_jwks(self) -> schemas.JWKSResponse:
        """
        Get the JWKS requried for authentication.
        See here for more details: https://auth0.com/docs/tokens/references/jwks-properties
        """
        rsa_obj = RSA.import_key(self.public_key)

        jwks_obj = schemas.JWKSResponse(
            alg=self.jwt_algorithm,
            kty=self.jwt_algorithm_family,
            use="sig",
            x5c=[self.public_key],
            n=INT_TO_BIG_ENDIAN(rsa_obj.n),
            e=INT_TO_BIG_ENDIAN(rsa_obj.e)
        )

        return jwks_obj
